from enum import Enum
from typing import Optional

# ---------------------------------------------------------------------------


class SRUOperation(str, Enum):
    """SRU operation"""

    def __str__(self) -> str:
        return self.value

    #: A ``explain`` operation
    EXPLAIN = "explain"
    """A ``explain`` operation"""

    SEARCH_RETRIEVE = "searchRetrieve"
    """A ``searchRetrieve`` operation"""

    SCAN = "scan"
    """A ``scan`` operation"""


class SRUQueryType(str, Enum):
    def __str__(self) -> str:
        return self.value

    CQL = "cql"
    """shorthand queryType identifier for CQL"""

    SEARCH_TERMS = "searchTerms"


class SRURecordPacking(str, Enum):
    """SRU 2.0 record packing."""

    def __str__(self) -> str:
        return self.value

    PACKED = "packed"
    """The client requests that the server should supply records strictly
    according to the requested schema."""

    UNPACKED = "unpacked"
    """The server is free to allow the location of application data to vary
    within the record."""


class SRURecordXmlEscaping(str, Enum):
    """SRU Record XML escaping (or record packing in SRU 1.2)."""

    def __str__(self) -> str:
        return self.value

    XML = "xml"
    """XML record packing"""

    STRING = "string"
    """String record packing"""


class SRURenderBy(str, Enum):
    """SRU Record XML escaping."""

    def __str__(self) -> str:
        return self.value

    CLIENT = "client"
    """The client requests that the server simply return this URL in
    the response, in the href attribute of the xml-stylesheet
    processing instruction before the response xml."""

    SERVER = "server"
    """The client requests that the server format the response
    according to the specified stylesheet, assuming the default SRU
    response schema as input to the stylesheet."""


class SRUResultCountPrecision(str, Enum):
    """(SRU 2.0) Indicate the accuracy of the result count reported
    by total number of records that matched the query."""

    def __str__(self) -> str:
        return self.value

    EXACT = "exact"
    """The server guarantees that the reported number of records is accurate."""

    UNKNOWN = "unknown"
    """The server has no idea what the result count is, and does not want to
    venture an estimate."""

    ESTIMATE = "estimate"
    """The server does not know the result set count, but offers an estimate."""

    MAXIMUM = "maximum"
    """The value supplied is an estimate of the maximum possible count that the
    result set will attain."""

    MINIMUM = "minimum"
    """The server does not know the result count but guarantees that it is at
    least this large."""

    CURRENT = "current"
    """The value supplied is an estimate of the count at the time the response
    was sent, however the result set may continue to grow."""


# ---------------------------------------------------------------------------


class SRUVersion(str, Enum):
    """SRU version"""

    def __new__(cls, major: int, minor: int):
        obj = str.__new__(cls, f"{major}.{minor}")
        obj._value_ = f"{major}.{minor}"
        obj.major = major
        obj.minor = minor
        return obj

    major: int
    minor: int

    @property
    def version_number(self) -> int:
        return (self.major << 16) | self.minor

    @property
    def version_string(self) -> str:
        return self.value

    def __str__(self) -> str:
        return self.value

    VERSION_1_1 = (1, 1)
    VERSION_1_2 = (1, 2)
    VERSION_2_0 = (2, 0)


# ---------------------------------------------------------------------------
# Diagnostics

SRU_DIAGNOSTIC_RECORD_SCHEMA = "info:srw/schema/1/diagnostics-v1.1"
SRU_DIAGNOSTIC_URI_PREFIX = "info:srw/diagnostic/1/"


class SRUDiagnostics(str, Enum):
    """Constants for SRU diagnostics

    See also:
        * SRU Diagnostics: http://www.loc.gov/standards/sru/diagnostics/
        * SRU Diagnostics List: http://www.loc.gov/standards/sru/diagnostics/diagnosticsList.html
    """

    def __new__(cls, nr: int, category: str, description: str):
        obj = str.__new__(cls, f"{SRU_DIAGNOSTIC_URI_PREFIX}{nr}")
        obj._value_ = f"{SRU_DIAGNOSTIC_URI_PREFIX}{nr}"
        obj.nr = nr
        obj.category = category
        obj.description = description
        obj.__doc__ = description
        return obj

    nr: int
    category: str
    description: str

    # fmt: off

    # general diagnostics
    GENERAL_SYSTEM_ERROR = (1, "general", "General system error")
    SYSTEM_TEMPORARILY_UNAVAILABLE = (2, "general", "System temporarily unavailable")
    AUTHENTICATION_ERROR = (3, "general", "Authentication error")
    UNSUPPORTED_OPERATION = (4, "general", "Unsupported operation")
    UNSUPPORTED_VERSION = (5, "general", "Unsupported version")
    UNSUPPORTED_PARAMETER_VALUE = (6, "general", "Unsupported parameter value")
    MANDATORY_PARAMETER_NOT_SUPPLIED = (7, "general", "Mandatory parameter not supplied")
    UNSUPPORTED_PARAMETER = (8, "general", "Unsupported Parameter")
    DATABASE_DOES_NOT_EXIST = (235, "general", "Database does not exist")

    # diagnostics relating to CQL
    QUERY_SYNTAX_ERROR = (10, "cql", "Query syntax erro")
    TOO_MANY_CHARACTERS_IN_QUERY = (12, "cql", "Too many characters in query")
    INVALID_OR_UNSUPPORTED_USE_OF_PARENTHESES = (13, "cql", "Invalid or unsupported use of parentheses")
    INVALID_OR_UNSUPPORTED_USE_OF_QUOTES = (14, "cql", "Invalid or unsupported use of quotes")
    UNSUPPORTED_CONTEXT_SET = (15, "cql", "Unsupported context set")
    UNSUPPORTED_INDEX = (16, "cql", "Unsupported index")
    UNSUPPORTED_COMBINATION_OF_INDEXES = (18, "cql", "Unsupported combination of indexes")
    UNSUPPORTED_RELATION = (19, "cql", "Unsupported relation")
    UNSUPPORTED_RELATION_MODIFIER = (20, "cql", "Unsupported relation modifier")
    UNSUPPORTED_COMBINATION_OF_RELATION_MODIFERS = (21, "cql", "Unsupported combination of relation modifers")
    UNSUPPORTED_COMBINATION_OF_RELATION_AND_INDEX = (22, "cql", "Unsupported combination of relation and index")
    TOO_MANY_CHARACTERS_IN_TERM = (23, "cql", "Too many characters in term")
    UNSUPPORTED_COMBINATION_OF_RELATION_AND_TERM = (24, "cql", "Unsupported combination of relation and term")
    NON_SPECIAL_CHARACTER_ESCAPED_IN_TERM = (26, "cql", "Non special character escaped in term")
    EMPTY_TERM_UNSUPPORTED = (27, "cql", "Empty term unsupported")
    MASKING_CHARACTER_NOT_SUPPORTED = (28, "cql", "Masking character not supported")
    MASKED_WORDS_TOO_SHORT = (29, "cql", "Masked words too short")
    TOO_MANY_MASKING_CHARACTERS_IN_TERM = (30, "cql", "Too many masking characters in term")
    ANCHORING_CHARACTER_NOT_SUPPORTED = (31, "cql", "Anchoring character not supported")
    ANCHORING_CHARACTER_IN_UNSUPPORTED_POSITION = (32, "cql", "Anchoring character in unsupported position")
    COMBINATION_OF_PROXIMITY_ADJACENCY_AND_MASKING_CHARACTERS_NOT_SUPPORTED = (33, "cql", "Combination of proximity/adjacency and masking characters not supported")
    COMBINATION_OF_PROXIMITY_ADJACENCY_AND_ANCHORING_CHARACTERS_NOT_SUPPORTED = (34, "cql", "Combination of proximity/adjacency and anchoring characters not supported")
    TERM_CONTAINS_ONLY_STOPWORDS = (35, "cql", "Term contains only stopwords")
    TERM_IN_INVALID_FORMAT_FOR_INDEX_OR_RELATION = (36, "cql", "Term in invalid format for index or relation")
    UNSUPPORTED_BOOLEAN_OPERATOR = (37, "cql", "Unsupported boolean operator")
    TOO_MANY_BOOLEAN_OPERATORS_IN_QUERY = (38, "cql", "Too many boolean operators in query")
    PROXIMITY_NOT_SUPPORTED = (39, "cql", "Proximity not supporte")
    UNSUPPORTED_PROXIMITY_RELATION = (40, "cql", "Unsupported proximity relation")
    UNSUPPORTED_PROXIMITY_DISTANCE = (41, "cql", "Unsupported proximity distance")
    UNSUPPORTED_PROXIMITY_UNIT = (42, "cql", "Unsupported proximity unit")
    UNSUPPORTED_PROXIMITY_ORDERING = (43, "cql", "Unsupported proximity ordering")
    UNSUPPORTED_COMBINATION_OF_PROXIMITY_MODIFIERS = (44, "cql", "Unsupported combination of proximity modifiers")
    UNSUPPORTED_BOOLEAN_MODIFIER = (46, "cql", "Unsupported boolean modifier")
    CANNOT_PROCESS_QUERY_REASON_UNKNOWN = (47, "cql", "Cannot process query; reason unknown")
    QUERY_FEATURE_UNSUPPORTED = (48, "cql", "Query feature unsupported")
    MASKING_CHARACTER_IN_UNSUPPORTED_POSITION = (49, "cql", "Masking character in unsupported position")

    # diagnostics relating to result sets
    RESULT_SETS_NOT_SUPPORTED = (50, "result sets", "Result sets not supported")
    RESULT_SET_DOES_NOT_EXIST = (51, "result sets", "Result set does not exist")
    RESULT_SET_TEMPORARILY_UNAVAILABLE = (52, "result sets", "Result set temporarily unavailable")
    RESULT_SETS_ONLY_SUPPORTED_FOR_RETRIEVAL = (53, "result sets", "Result sets only supported for retrieval")
    COMBINATION_OF_RESULT_SETS_WITH_SEARCH_TERMS_NOT_SUPPORTED = (55, "result sets", "Combination of result sets with search terms not supported")
    RESULT_SET_CREATED_WITH_UNPREDICTABLE_PARTIAL_RESULTS_AVAILABLE = (58, "result sets", "Result set created with unpredictable partial results available")
    RESULT_SET_CREATED_WITH_VALID_PARTIAL_RESULTS_AVAILABLE = (59, "result sets", "Result set created with valid partial results available")
    RESULT_SET_NOT_CREATED_TOO_MANY_MATCHING_RECORDS = (60, "result sets", "Result set not created: too many matching records")

    # diagnostics relating to records
    FIRST_RECORD_POSITION_OUT_OF_RANGE = (61, "records", "First record position out of range")
    RECORD_TEMPORARILY_UNAVAILABLE = (64, "records", "Record temporarily unavailable")
    RECORD_DOES_NOT_EXIST = (65, "records", "Record does not exist")
    UNKNOWN_SCHEMA_FOR_RETRIEVAL = (66, "records", "Unknown schema for retrieval")
    RECORD_NOT_AVAILABLE_IN_THIS_SCHEMA = (67, "records", "Record not available in this schema")
    NOT_AUTHORISED_TO_SEND_RECORD = (68, "records", "Not authorised to send record")
    NOT_AUTHORISED_TO_SEND_RECORD_IN_THIS_SCHEMA = (69, "records", "Not authorised to send record in this schema")
    RECORD_TOO_LARGE_TO_SEND = (70, "records", "Record too large to send")
    UNSUPPORTED_XML_ESCAPING_VALUE = (71, "records", "Unsupported record packing")
    XPATH_RETRIEVAL_UNSUPPORTED = (72, "records", "XPath retrieval unsupported")
    XPATH_EXPRESSION_CONTAINS_UNSUPPORTED_FEATURE = (73, "records", "XPath expression contains unsupported feature")
    UNABLE_TO_EVALUATE_XPATH_EXPRESSION = (74, "records", "Unable to evaluate XPath expression")

    # diagnostics relating to sorting
    SORT_NOT_SUPPORTED = (80, "sorting", "Sort not supported")
    UNSUPPORTED_SORT_SEQUENCE = (82, "sorting", "Unsupported sort sequence")
    TOO_MANY_RECORDS_TO_SORT = (83, "sorting", "Too many records to sort")
    TOO_MANY_SORT_KEYS_TO_SORT = (84, "sorting", "Too many sort keys to sort")
    CANNOT_SORT_INCOMPATIBLE_RECORD_FORMATS = (86, "sorting", "Cannot sort: incompatible record format")
    UNSUPPORTED_SCHEMA_FOR_SORT = (87, "sorting", "Unsupported schema for sort")
    UNSUPPORTED_PATH_FOR_SORT = (88, "sorting", "Unsupported path for sort")
    PATH_UNSUPPORTED_FOR_SCHEMA = (89, "sorting", "Path unsupported for schema")
    UNSUPPORTED_DIRECTION = (90, "sorting", "Unsupported direction")
    UNSUPPORTED_CASE = (91, "sorting", "Unsupported case")
    UNSUPPORTED_MISSING_VALUE_ACTION = (92, "sorting", "Unsupported missing value action")
    SORT_ENDED_DUE_TO_MISSING_VALUE = (93, "sorting", "Sort ended due to missing value ")
    SORT_SPEC_INCLUDED_BOTH_IN_QUERY_AND_PROTOCOL_QUERY_PREVAILS = (94, "sorting", "Sort spec included both in query and protocol: query prevails")
    SORT_SPEC_INCLUDED_BOTH_IN_QUERY_AND_PROTOCOL_PROTOCOL_PREVAILS = (95, "sorting", "Sort spec included both in query and protocol: protocol prevails")
    SORT_SPEC_INCLUDED_BOTH_IN_QUERY_AND_PROTOCOL_ERROR = (96, "sorting", "Sort spec included both in query and protocol: error")

    # diagnostics relating to stylesheets
    STYLESHEETS_NOT_SUPPORTED = (110, "stylesheets", "Stylesheets not supported")
    UNSUPPORTED_STYLESHEET = (111, "stylesheets", "Unsupported stylesheet")

    # diagnostics relating to scan
    RESPONSE_POSITION_OUT_OF_RANGE = (120, "scan", "Response position out of range")
    TOO_MANY_TERMS_REQUESTED = (121, "scan", "Too many terms requested")

    # fmt: on

    @classmethod
    def get_by_uri(cls, uri: str) -> Optional["SRUDiagnostics"]:
        if not uri:
            return None
        for _, member in cls.__members__.items():
            if member.value == uri:
                return member
        return None


# ---------------------------------------------------------------------------


RESPONSE_ENCODING = "utf-8"
RESPONSE_CONTENT_TYPE = "application/xml"

PARAM_EXTENSION_PREFIX = "x-"


class SRUParam(str, Enum):
    def __str__(self) -> str:
        return self.value

    # general / explain related parameter names
    OPERATION = "operation"
    VERSION = "version"
    STYLESHEET = "stylesheet"
    RENDER_BY = "renderedBy"
    HTTP_ACCEPT = "httpAccept"
    RESPONSE_TYPE = "responseType"

    # searchRetrieve related parameter names
    QUERY = "query"
    QUERY_TYPE = "queryType"
    START_RECORD = "startRecord"
    MAXIMUM_RECORDS = "maximumRecords"
    RECORD_XML_ESCAPING = "recordXMLEscaping"
    RECORD_PACKING = "recordPacking"
    RECORD_SCHEMA = "recordSchema"
    RECORD_XPATH = "recordXPath"
    RESULT_SET_TTL = "resultSetTTL"
    SORT_KEYS = "sortKeys"

    # scan related parameter names
    SCAN_CLAUSE = "scanClause"
    RESPONSE_POSITION = "responsePosition"
    MAXIMUM_TERMS = "maximumTerms"

    X_UNLIMITED_RESULTSET = "x-unlimited-resultset"
    X_UNLIMITED_TERMLIST = "x-unlimited-termlist"
    X_INDENT_RESPONSE = "x-indent-response"


class SRUParamValue(str, Enum):
    def __str__(self) -> str:
        return self.value

    # operations
    OP_EXPLAIN = "explain"
    OP_SCAN = "scan"
    OP_SEARCH_RETRIEVE = "searchRetrieve"
    VERSION_1_1 = "1.1"
    VERSION_1_2 = "1.2"

    # various parameter values
    RECORD_XML_ESCAPING_XML = "xml"
    RECORD_XML_ESCAPING_STRING = "string"
    RECORD_PACKING_PACKED = "packed"
    RECORD_PACKING_UNPACKED = "unpacked"
    RENDER_BY_CLIENT = "client"
    RENDER_BY_SERVER = "server"


# ---------------------------------------------------------------------------



import io
import xml.etree.ElementTree as ET
import xml.sax.saxutils
from collections import deque
from contextlib import contextmanager
from enum import Enum
from enum import auto
from typing import Deque
from typing import Union
from xml.sax import SAXException
from xml.sax.handler import ContentHandler
from xml.sax.saxutils import XMLGenerator

import cql
from lxml import etree
from lxml.sax import saxify

from ..constants import RESPONSE_ENCODING
from ..constants import SRURecordXmlEscaping

# ---------------------------------------------------------------------------


class SRUXMLStreamWriter(ContentHandler):
    class IndentingState(Enum):
        SEEN_NOTHING = auto()
        SEEN_ELEMENT = auto()
        SEEN_DATA = auto()

    def __init__(
        self,
        output_stream: io.TextIOBase,
        record_escaping: SRURecordXmlEscaping,
        indent: int = -1,
        encoding: str = RESPONSE_ENCODING,
        short_empty_elements: bool = False,
    ) -> None:
        super().__init__()
        self.record_escaping = record_escaping
        self.writing_record = False
        # TODO: SRURecordXmlEscaping.STRING
        # https://www.loc.gov/standards/sru/sru-1-1.html#packing
        # if self.writing_record and self.record_escaping == SRURecordXmlEscaping.STRING:
        #     content = escape(content)

        self.depth = 0
        self.indent = indent
        self.indent_state = SRUXMLStreamWriter.IndentingState.SEEN_NOTHING
        self.indent_state_stack: Deque[SRUXMLStreamWriter.IndentingState] = deque()

        self.output_stream = self.output_stream_raw = output_stream

        if self.record_escaping == SRURecordXmlEscaping.STRING:
            # output_stream.__class__ ?
            class SRURecordXmlEscapingStream(io.TextIOBase):
                def __init__(self, writer) -> None:
                    super().__init__()
                    self.writer = writer

                def write(self, __s: str) -> int:
                    if self.writer.writing_record:
                        __s = xml.sax.saxutils.escape(__s)
                    return self.writer.output_stream.write(__s)

                def flush(self) -> None:
                    return self.writer.output_stream.flush()

            # FIXME: is that stable, even required?
            # self.output_stream = xml.sax.saxutils._gettextwriter(
            #     self.output_stream, encoding
            # )
            xml_output_stream: Union[
                SRURecordXmlEscapingStream, io.TextIOBase
            ] = SRURecordXmlEscapingStream(self)

        else:
            xml_output_stream = self.output_stream

        self.xmlwriter = XMLGenerator(
            xml_output_stream,
            encoding=encoding,
            short_empty_elements=short_empty_elements,
        )

    # ----------------------------------------------------

    def _should_do_indent_stuff(self):
        if not self.writing_record:
            return True
        if self.record_escaping != SRURecordXmlEscaping.STRING:
            return True
        return False

    def onStartElement(self):
        if self._should_do_indent_stuff():
            self.indent_state_stack.append(
                SRUXMLStreamWriter.IndentingState.SEEN_ELEMENT
            )
            self.indent_state = SRUXMLStreamWriter.IndentingState.SEEN_NOTHING
            if self.depth > 0:
                self.xmlwriter.characters("\n")
            self.doIndent()
            self.depth += 1

    def onEndElement(self):
        if self._should_do_indent_stuff():
            self.depth -= 1
            if self.indent_state == SRUXMLStreamWriter.IndentingState.SEEN_ELEMENT:
                self.xmlwriter.characters("\n")
                self.doIndent()
            self.indent_state = self.indent_state_stack.pop()

    def onEmptyElement(self):
        if self._should_do_indent_stuff:
            self.indent_state = SRUXMLStreamWriter.IndentingState.SEEN_ELEMENT
            if self.depth > 0:
                self.xmlwriter.characters("\n")
            self.doIndent()

    def doIndent(self):
        if self.depth > 0:
            self.xmlwriter.characters(" " * self.depth * self.indent)

    # ----------------------------------------------------

    def startRecord(self):
        if self.writing_record:
            raise ValueError("was already writing record")
        self.xmlwriter._flush()  # or call on my stream variable?
        # force writer to close/finish any pending start or end elements
        self.xmlwriter._finish_pending_start_element()
        self.writing_record = True

    def endRecord(self):
        if not self.writing_record:
            raise ValueError("was not writing record")
        # force writer to close/finish any pending start or end elements
        self.xmlwriter._finish_pending_start_element()
        self.xmlwriter._flush()  # or call on my stream variable?
        self.writing_record = False

    # ----------------------------------------------------
    # ContentHandler methods

    def setDocumentLocator(self, locator):
        self.xmlwriter.setDocumentLocator(locator)

    def startPrefixMapping(self, prefix, uri):
        self.xmlwriter.startPrefixMapping(prefix, uri)

    def endPrefixMapping(self, prefix):
        self.xmlwriter.endPrefixMapping(prefix)

    def processingInstruction(self, target, data):
        self.xmlwriter.processingInstruction(target, data)

    def startDocument(self):
        self.xmlwriter.startDocument()
        # if self.indent > 0:
        #     self.xmlwriter.characters("\n")

    def endDocument(self):
        self.xmlwriter.endDocument()

    def startElement(self, name, attrs=None):
        if self.indent > 0:
            self.onStartElement()
        if attrs is None:
            attrs = dict()
        self.xmlwriter.startElement(name, attrs)

    def endElement(self, name):
        if self.indent > 0:
            self.onEndElement()
        self.xmlwriter.endElement(name)

    def startElementNS(self, name, qname=None, attrs=None):
        if self.indent > 0:
            self.onStartElement()
        if attrs is None:
            attrs = dict()
        else:
            # small helper to set None-namespace for attributes
            # that did not supply them
            if not all(isinstance(key, tuple) for key in attrs.keys()):
                attrs_copy = dict()
                for key, value in attrs.items():
                    if not isinstance(key, tuple):
                        key = (None, key)
                    attrs_copy[key] = value
                attrs = attrs_copy
        self.xmlwriter.startElementNS(name, qname, attrs)

    def endElementNS(self, name, qname=None):
        if self.indent > 0:
            self.onEndElement()
        self.xmlwriter.endElementNS(name, qname)

    def characters(self, content):
        if self.indent > 0:
            self.indent_state = SRUXMLStreamWriter.IndentingState.SEEN_DATA
        self.xmlwriter.characters(content)

    def ignorableWhitespace(self, whitespace):
        self.xmlwriter.ignorableWhitespace(whitespace)

    def skippedEntity(self, name):
        self.xmlwriter.skippedEntity(name)

    # ----------------------------------------------------

    def writeXCQL(self, query: cql.CQLQuery, search_retrieve_mode: bool):
        # HACK: Parsing the XCQL to serialize is wasting resources.
        # Alternative would be to serialize to XCQL from CQLNode, but
        # I'm not yet enthusiastic on writing the serializer myself.

        class XCQLHandler(ContentHandler):
            def __init__(self, writer):
                super().__init__()
                self.writer = writer

            def startElementNS(self, name, qname, attrs):
                if not search_retrieve_mode and qname == "searchClause":
                    return
                self.writer.startElementNS(name, qname, attrs)

            def endElementNS(self, name, qname):
                if not search_retrieve_mode and qname == "searchClause":
                    return
                self.writer.endElementNS(name, qname)

            def characters(self, content):
                if not content or content.isspace():
                    return
                self.writer.characters(content)

        try:
            # tree = query.toXCQL()
            # content = query.toXCQLString()
            tree = query.root.toXCQL()
            content = ET.tostring(tree)
            tree = etree.fromstring(content)
            handler = XCQLHandler(self)
            saxify(tree, handler)
        except Exception as ex:
            raise SAXException("serializing xcql failed") from ex

    # ----------------------------------------------------

    @contextmanager
    def prefix(self, prefix, uri):
        self.startPrefixMapping(prefix, uri)
        yield
        self.endPrefixMapping(prefix)

    @contextmanager
    def element(self, name, namespace=None, attrs=None):
        self.startElementNS((namespace, name), attrs=attrs)
        yield
        self.endElementNS((namespace, name))

    def elementcontent(self, name, content=None, namespace=None, attrs=None):
        self.startElementNS((namespace, name), attrs=attrs)
        self.characters(content)
        self.endElementNS((namespace, name))

    @contextmanager
    def record(self):
        self.startRecord()
        yield
        self.endRecord()


# ---------------------------------------------------------------------------


def copy_XML_into_writer(writer: ContentHandler, xml: Union[bytes, str]):
    class XMLCopyHandler(ContentHandler):
        def __init__(self, writer):
            super().__init__()
            self.writer = writer

        def startPrefixMapping(self, prefix, uri):
            self.writer.startPrefixMapping(prefix, uri)

        def endPrefixMapping(self, prefix):
            self.writer.endPrefixMapping(prefix)

        def startElementNS(self, name, qname, attrs):
            self.writer.startElementNS(name, qname, attrs)

        def endElementNS(self, name, qname):
            self.writer.endElementNS(name, qname)

        def characters(self, content):
            # if not content or content.isspace():
            #     return
            self.writer.characters(content)

    try:
        tree = etree.fromstring(xml)
        handler = XMLCopyHandler(writer)
        saxify(tree, handler)
    except Exception as ex:
        raise SAXException("serializing xml failed") from ex


class XMLStreamWriterHelper(ContentHandler):
    def __init__(self, xmlwriter: ContentHandler) -> None:
        super().__init__()
        self.xmlwriter = xmlwriter

        # unwrap to avoid uneccessary call chains
        if (
            isinstance(self.xmlwriter, XMLStreamWriterHelper)
            and self.xmlwriter.__class__ == XMLStreamWriterHelper
        ):
            self.xmlwriter = self.xmlwriter.xmlwriter

    # ----------------------------------------------------
    # ContentHandler methods

    def setDocumentLocator(self, locator):
        self.xmlwriter.setDocumentLocator(locator)

    def startPrefixMapping(self, prefix, uri):
        self.xmlwriter.startPrefixMapping(prefix, uri)

    def endPrefixMapping(self, prefix):
        self.xmlwriter.endPrefixMapping(prefix)

    def processingInstruction(self, target, data):
        self.xmlwriter.processingInstruction(target, data)

    def startDocument(self):
        self.xmlwriter.startDocument()

    def endDocument(self):
        self.xmlwriter.endDocument()

    def startElement(self, name, attrs=None):
        if attrs is None:
            attrs = dict()
        self.xmlwriter.startElement(name, attrs)

    def endElement(self, name):
        self.xmlwriter.endElement(name)

    def startElementNS(self, name, qname=None, attrs=None):
        if attrs is None:
            attrs = dict()
        else:
            # small helper to set None-namespace for attributes
            # that did not supply them
            if not all(isinstance(key, tuple) for key in attrs.keys()):
                attrs_copy = dict()
                for key, value in attrs.items():
                    if not isinstance(key, tuple):
                        key = (None, key)
                    attrs_copy[key] = value
                attrs = attrs_copy
        self.xmlwriter.startElementNS(name, qname, attrs)

    def endElementNS(self, name, qname=None):
        self.xmlwriter.endElementNS(name, qname)

    def characters(self, content):
        self.xmlwriter.characters(content)

    def ignorableWhitespace(self, whitespace):
        self.xmlwriter.ignorableWhitespace(whitespace)

    def skippedEntity(self, name):
        self.xmlwriter.skippedEntity(name)

    # ----------------------------------------------------

    def writeXML(self, xml: Union[bytes, str]):
        copy_XML_into_writer(self, xml)

    def writeXMLdocument(self, xmldoc: ET.Element):
        try:
            content = ET.tostring(xmldoc)
            self.writeXML(content)
        except SAXException:
            raise
        except Exception as ex:
            raise SAXException("serializing xmldoc failed") from ex

    # ----------------------------------------------------
    # contextmanagers + SRUXMLStreamWriter methods

    @contextmanager
    def prefix(self, prefix, uri):
        self.startPrefixMapping(prefix, uri)
        yield
        self.endPrefixMapping(prefix)

    @contextmanager
    def element(self, name, namespace=None, attrs=None):
        self.startElementNS((namespace, name), attrs=attrs)
        yield
        self.endElementNS((namespace, name))

    def elementcontent(self, name, content=None, namespace=None, attrs=None):
        self.startElementNS((namespace, name), attrs=attrs)
        self.characters(content)
        self.endElementNS((namespace, name))

    def startRecord(self):
        if isinstance(self.xmlwriter, SRUXMLStreamWriter):
            self.xmlwriter.startRecord()

    def endRecord(self):
        if isinstance(self.xmlwriter, SRUXMLStreamWriter):
            self.xmlwriter.endRecord()

    @contextmanager
    def record(self):
        self.startRecord()
        yield
        self.endRecord()


# ---------------------------------------------------------------------------



import logging
import re
from abc import ABCMeta
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

import cql
from werkzeug.wrappers import Request

from ..constants import PARAM_EXTENSION_PREFIX
from ..constants import SRUDiagnostics
from ..constants import SRUOperation
from ..constants import SRUParam
from ..constants import SRUParamValue
from ..constants import SRUQueryType
from ..constants import SRURecordPacking
from ..constants import SRURecordXmlEscaping
from ..constants import SRURenderBy
from ..constants import SRUVersion
from ..diagnostic import SRUDiagnostic
from ..diagnostic import SRUDiagnosticList
from ..exception import SRUException
from ..queryparser import CQLQueryParser
from ..queryparser import SRUQuery
from ..queryparser import SRUQueryParserRegistry
from .auth import SRUAuthenticationInfo
from .auth import SRUAuthenticationInfoProvider
from .config import SRUServerConfig

T = TypeVar("T")

LOGGER = logging.getLogger("__name__")

QUERY_TYPE_ALLOWED_CHARS = re.compile(r"[a-zA-Z0-9][a-zA-Z0-9_-]*")


# ---------------------------------------------------------------------------


class SRURequest(metaclass=ABCMeta):
    """Provides information about a SRU request."""

    @abstractmethod
    def get_operation(self) -> SRUOperation:
        """Get the ``operation`` parameter of this request. Available
        for **explain**, **searchRetrieve** and **scan** requests.
        """

    @abstractmethod
    def get_version(self) -> SRUVersion:
        """Get the **version** parameter of this request. Available
        for **explain**, **searchRetrieve** and **scan** requests.
        """

    def is_version(self, version: SRUVersion) -> bool:
        """Check if this request is of a specific version.

        Args:
            version: the version to check

        Returns:
            bool: ``True`` if this request is in the requested
                version, ``False`` otherwise
        """
        if version is None:
            raise Type("version is None")
        return self.get_version() == version

    def is_version_between(self, min: SRUVersion, max: SRUVersion) -> bool:
        """Check if version of this request is at least `min` and
        at most `max`.

        Args:
            min: the minimum version
            max: the maximum version

        Returns:
            bool: ``True`` if this request is in the requested
                version, ``False`` otherwise
        """
        if min is None:
            raise TypeError("min is None")
        if max is None:
            raise TypeError("max is None")
        if min.version_number > max.version_number:
            raise ValueError("min > max")
        version = self.get_version()
        return (
            version.version_number >= min.version_number
            and version.version_number <= max.version_number
        )

    @abstractmethod
    def get_record_xml_escaping(self) -> SRURecordXmlEscaping:
        """Get the **recordXmlEscpaing** (SRU 2.0) or **recordPacking**
        (SRU 1.1 and SRU 1.2) parameter of this request. Only
        available for **explain** and **searchRetrieve** requests.

        Returns:
            SRURecordXmlEscaping: the record XML escaping method
        """

    @abstractmethod
    def get_record_packing(self) -> SRURecordPacking:
        """Get the **recordPacking** (SRU 2.0) parameter of this
        request. Only available for **searchRetrieve** requests.

        Returns:
            SRURecordPacking: the record packing method
        """

    @abstractmethod
    def get_query(self) -> Optional[SRUQuery[Any]]:
        """Get the **query** parameter of this request. Only available
        for **searchRetrieve** requests.

        Returns:
            SRUQuery[Any]: an `SRUQuery` instance tailored for the
                used queryType or `None` if not a **searchRetrieve**
                request
        """

    # TODO: required; pythonic?
    # def get_query(self, type: Type[T]) -> Optional[SRUQuery[T]]:

    def get_query_type(self) -> Optional[str]:
        """Get the **queryType** parameter of this request. Only
        available for **searchRetrieve** requests.

        Returns:
            str: the queryType of the parsed query or `None` if not a
                **searchRetrieve** request
        """
        query = self.get_query()
        if query is None:
            return None
        return query.query_type

    def is_query_type(self, query_type: str) -> bool:
        """Check if the request was made with the given queryType.
        Only available for **searchRetrieve** requests.

        Args:
            query_type: the queryType to compare with

        Returns:
            bool: ``True`` if the queryType matches, ``False``
                otherwise
        """
        if query_type is None:
            return False
        return self.get_query_type() == query_type

    @abstractmethod
    def get_start_record(self) -> int:
        """Get the **startRecord** parameter of this request. Only
        available for **searchRetrieve** requests. If the client did
        not provide a value for the request, it is set to ``1``.

        Returns:
            int: the number of the start record
        """

    @abstractmethod
    def get_maximum_records(self) -> int:
        """Get the **maximumRecords** parameter of this request. Only
        available for **searchRetrieve** requests. If no value was
        supplied with the request, the server will automatically set
        a default value.

        Returns:
            int: the maximum number of records
        """

    @abstractmethod
    def get_record_schema_identifier(self) -> Optional[str]:
        """Get the record schema identifier derived from the
        **recordSchema** parameter of this request. Only available
        for **searchRetrieve** requests. If the request was send with
        the short record schema name, it will automatically expanded
        to the record schema identifier.

        Returns:
            str: the record schema identifier or `None` if no
                **recordSchema** parameter was supplied for this
                request
        """

    @abstractmethod
    def get_record_xpath(self) -> Optional[str]:
        """Get the **recordXPath** parameter of this request. Only
        available for **searchRetrieve** requests and version 1.1
        requests.

        Returns:
            str: the record XPath or `None` of no value was supplied
                for this request
        """

    @abstractmethod
    def get_resultSet_TTL(self) -> int:
        """Get the **resultSetTTL** parameter of this request. Only
        available for **searchRetrieve** requests.

        Returns:
            int: the result set TTL or ``-1`` if no value was
                supplied for this request
        """

    @abstractmethod
    def get_sortKeys(self) -> Optional[str]:
        """Get the **sortKeys** parameter of this request. Only
        available for **searchRetrieve** requests and version 1.1 requests.

        Returns:
            str: the record XPath or `None` of no value was supplied
                for this request
        """

    # TODO CQLQuery/CQLNode?
    @abstractmethod
    def get_scan_clause(self) -> Optional[cql.CQLQuery]:
        """Get the **scanClause** parameter of this request. Only
        available for **scan** requests.

        Returns:
            cql.CQLQuery: the parsed scan clause or `None` if not a
                **scan** request
        """

    @abstractmethod
    def get_response_position(self) -> int:
        """Get the **responsePosition** parameter of this request.
        Only available for **scan** requests. If the client did not
        provide a value for the request, it is set to ``1``.

        Returns:
            int: the response position
        """

    @abstractmethod
    def get_maximum_terms(self) -> int:
        """Get the **maximumTerms** parameter of this request.
        Available for any type of request.

        Returns:
            int: the maximum number of terms or ``-1`` if no value
                was supplied for this request
        """

    @abstractmethod
    def get_stylesheet(self) -> Optional[str]:
        """Get the **stylesheet** parameter of this request.
        Available for **explain**, **searchRetrieve** and **scan**
        requests.

        Returns:
            str: the stylesheet or `None` if no value was supplied
                for this request
        """

    @abstractmethod
    def get_renderBy(self) -> Optional[SRURenderBy]:
        """Get the **renderBy** parameter of this request.

        Returns:
            SRURenderBy: the renderBy parameter or `None` if no value
                was supplied for this request
        """

    @abstractmethod
    def get_response_type(self) -> Optional[str]:
        """(SRU 2.0) The request parameter **responseType**, paired
        with the Internet media type specified for the response (via
        either the httpAccept parameter or http accept header)
        determines the schema for the response.

        Returns:
            str: the value of the responeType request parameter or
                `None` if no value was supplied for this request
        """

    @abstractmethod
    def get_http_accept(self) -> Optional[str]:
        """(SRU 2.0) The request parameter **httpAccept** may be
        supplied to indicate the preferred format of the response.
        The value is an Internet media type.

        Returns:
            str: the value of the httpAccept request parameter or
                `None` if no value was supplied for
        """

    @abstractmethod
    def get_protocol_schema(self) -> str:
        """Get the protocol schema which was used of this request.
        Available for **explain**, **searchRetrieve** and **scan**
        requests.

        Returns:
            str: the protocol scheme
        """

    @abstractmethod
    def get_extra_request_data_names(self) -> List[str]:
        """Get the names of extra parameters of this request.
        Available for **explain**, **searchRetrieve** and **scan**
        requests.

        Returns:
            List[str]: a possibly empty list of parameter names
        """

    @abstractmethod
    def get_extra_request_data(self, name: str) -> Optional[str]:
        """Get the value of an extra parameter of this request.
        Available for **explain**, **searchRetrieve** and **scan**
        requests.

        Args:
            name: name of the extra parameter. Must be prefixed with
                ``x-``

        Returns:
            str: the value of the parameter of `None` of extra
                parameter with that name exists
        """


# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class ParameterInfo:
    class Parameter(str, Enum):
        STYLESHEET = "stylesheet"
        RENDER_BY = "render_by"
        HTTP_ACCEPT = "http_accept"
        RESPONSE_TYPE = "response_type"
        START_RECORD = "start_record"
        MAXIMUM_RECORDS = "maximum_records"
        RECORD_XML_ESCAPING = "record_xml_escaping"
        RECORD_PACKING = "record_packing"
        RECORD_SCHEMA = "record_schema"
        RECORD_XPATH = "record_xpath"
        RESULT_SET_TTL = "result_set_ttl"
        SORT_KEYS = "sort_keys"
        SCAN_CLAUSE = "scan_clause"
        RESPONSE_POSITION = "response_position"
        MAXIMUM_TERMS = "maximum_terms"

    # ----------------------------------------------------# ----------------------------------------------------

    parameter: Parameter
    mandatory: bool
    min: SRUVersion
    max: SRUVersion

    def name(self, version: SRUVersion) -> Optional[str]:
        if self.parameter == ParameterInfo.Parameter.STYLESHEET:
            return SRUParam.STYLESHEET
        if self.parameter == ParameterInfo.Parameter.RENDER_BY:
            return SRUParam.RENDER_BY
        if self.parameter == ParameterInfo.Parameter.HTTP_ACCEPT:
            return SRUParam.HTTP_ACCEPT
        if self.parameter == ParameterInfo.Parameter.RESPONSE_TYPE:
            return SRUParam.RESPONSE_TYPE
        if self.parameter == ParameterInfo.Parameter.START_RECORD:
            return SRUParam.START_RECORD
        if self.parameter == ParameterInfo.Parameter.MAXIMUM_RECORDS:
            return SRUParam.MAXIMUM_RECORDS
        if self.parameter == ParameterInfo.Parameter.RECORD_SCHEMA:
            return SRUParam.RECORD_SCHEMA
        if self.parameter == ParameterInfo.Parameter.RECORD_XPATH:
            return SRUParam.RECORD_XPATH
        if self.parameter == ParameterInfo.Parameter.RESULT_SET_TTL:
            return SRUParam.RESULT_SET_TTL
        if self.parameter == ParameterInfo.Parameter.SORT_KEYS:
            return SRUParam.SORT_KEYS
        if self.parameter == ParameterInfo.Parameter.SCAN_CLAUSE:
            return SRUParam.SCAN_CLAUSE
        if self.parameter == ParameterInfo.Parameter.RESPONSE_POSITION:
            return SRUParam.RESPONSE_POSITION
        if self.parameter == ParameterInfo.Parameter.MAXIMUM_TERMS:
            return SRUParam.MAXIMUM_TERMS

        if self.parameter == ParameterInfo.Parameter.RECORD_XML_ESCAPING:
            """
            'recordPacking' was renamed to 'recordXMLEscaping' in SRU 2.0.
            For library API treat 'recordPacking' parameter as 'recordPacking'
            for SRU 1.1 and SRU 1.2.
            """
            if version == SRUVersion.VERSION_2_0:
                return SRUParam.RECORD_XML_ESCAPING
            else:
                return SRUParam.RECORD_PACKING
        if self.parameter == ParameterInfo.Parameter.RECORD_PACKING:
            """
            'recordPacking' only exists in SRU 2.0; the old variant is
            handled by the case for RECORD_XML_ESCAPING
            """
            if version == SRUVersion.VERSION_2_0:
                return SRUParam.RECORD_PACKING
            else:
                return None

        raise ValueError(f"unknown ParameterInfo.Parameter? {self.parameter}")

    def is_for_version(self, version: SRUVersion) -> bool:
        return (
            self.min.version_number <= version.version_number
            and self.max.version_number >= version.version_number
        )


class ParameterInfoSets(Enum):
    EXPLAIN = [
        ParameterInfo(
            ParameterInfo.Parameter.STYLESHEET,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_1_2,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.RECORD_XML_ESCAPING,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_1_2,
        ),
    ]
    SCAN = [
        ParameterInfo(
            ParameterInfo.Parameter.STYLESHEET,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.HTTP_ACCEPT,
            False,
            SRUVersion.VERSION_2_0,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.SCAN_CLAUSE,
            True,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.RESPONSE_POSITION,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.MAXIMUM_TERMS,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
    ]
    SEARCH_RETRIEVE = [
        ParameterInfo(
            ParameterInfo.Parameter.STYLESHEET,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_1_2,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.HTTP_ACCEPT,
            False,
            SRUVersion.VERSION_2_0,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.RENDER_BY,
            False,
            SRUVersion.VERSION_2_0,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.RESPONSE_TYPE,
            False,
            SRUVersion.VERSION_2_0,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.START_RECORD,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.MAXIMUM_RECORDS,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.RECORD_XML_ESCAPING,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.RECORD_PACKING,
            False,
            SRUVersion.VERSION_2_0,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.RECORD_SCHEMA,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.RESULT_SET_TTL,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.RECORD_XPATH,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_1_2,
        ),
        ParameterInfo(
            ParameterInfo.Parameter.SORT_KEYS,
            False,
            SRUVersion.VERSION_1_1,
            SRUVersion.VERSION_2_0,
        ),
    ]

    @classmethod
    def for_operation(
        cls, operation: Optional[SRUOperation]
    ) -> Optional[List[ParameterInfo]]:
        if not operation:
            return None
        if operation == SRUOperation.EXPLAIN:
            return cls.EXPLAIN.value
        if operation == SRUOperation.SCAN:
            return cls.SCAN.value
        if operation == SRUOperation.SEARCH_RETRIEVE:
            return cls.SEARCH_RETRIEVE.value
        # actually cannot happen
        return None


# ---------------------------------------------------------------------------

DEFAULT_START_RECORD = 1
DEFAULT_RESPONSE_POSITION = 1


class SRURequestImpl(SRUDiagnosticList, SRURequest):
    def __init__(
        self,
        config: SRUServerConfig,
        query_parsers: SRUQueryParserRegistry,
        request: Request,
        authentication_info_provider: Optional[SRUAuthenticationInfoProvider] = None,
    ):
        self.config = config
        self.query_parsers = query_parsers
        self.authentication_info_provider = authentication_info_provider
        self.authentication_info: Optional[SRUAuthenticationInfo] = None

        self.request = request

        self.diagnostics: List[SRUDiagnostic] = list()

        # NOTE: set default to EXPLAIN
        # (usually correctly set when parameters validated but operations
        # expect some value to be set, not None allowed)
        # FIXME: default value version None handling?
        self.operation: SRUOperation = SRUOperation.EXPLAIN
        self.version: Optional[SRUVersion] = None

        self.response_type: Optional[str] = None
        self.http_accept: Optional[str] = None

        self.record_xml_escaping: Optional[SRURecordXmlEscaping] = None
        self.record_packing: Optional[SRURecordPacking] = None
        self.renderBy: Optional[SRURenderBy] = None
        self.stylesheet: Optional[str] = None

        self.query: Optional[SRUQuery[Any]] = None

        self.start_record = DEFAULT_START_RECORD
        self.maximum_records = -1
        self.response_position = DEFAULT_RESPONSE_POSITION
        self.maximum_terms = -1
        self.record_schema_identifier: Optional[str] = None
        self.record_xpath: Optional[str] = None
        self.resultSet_TTL = -1
        self.sortKeys: Optional[str] = None
        self.scan_clause: Optional[cql.CQLQuery] = None

    # ----------------------------------------------------

    def get_request(self) -> Request:
        return self.request

    def get_operation(self) -> SRUOperation:
        return self.operation

    def get_version(self) -> SRUVersion:
        if self.version is not None:
            return self.version
        return self.config.default_version

    def get_authentication(self) -> Optional[SRUAuthenticationInfo]:
        return self.authentication_info

    def get_authentication_subject(self) -> Optional[str]:
        if not self.authentication_info:
            return None
        return self.authentication_info.subject

    # ----------------------------------------------------

    def get_query(self) -> Optional[SRUQuery[Any]]:
        return self.query

    def get_record_xml_escaping(self) -> SRURecordXmlEscaping:
        if self.record_xml_escaping is not None:
            return self.record_xml_escaping
        return self.config.default_record_xml_escaping

    def get_record_packing(self) -> SRURecordPacking:
        if self.record_packing is not None:
            return self.record_packing
        return self.config.default_record_packing

    def get_start_record(self) -> int:
        return self.start_record

    def get_maximum_records(self) -> int:
        if self.config.allow_override_maximum_records and self.get_extra_request_data(
            SRUParam.X_UNLIMITED_RESULTSET
        ):
            return -1
        if self.maximum_records == -1:
            return self.config.number_of_records
        if self.maximum_records > self.config.maximum_records:
            return self.config.maximum_records
        return self.maximum_records

    def get_record_schema_identifier(self) -> Optional[str]:
        return self.record_schema_identifier

    def get_record_xpath(self) -> Optional[str]:
        return self.record_xpath

    def get_resultSet_TTL(self) -> int:
        return self.resultSet_TTL

    def get_sortKeys(self) -> Optional[str]:
        return self.sortKeys

    def get_scan_clause(self) -> Optional[cql.CQLQuery]:
        return self.scan_clause

    def get_response_position(self) -> int:
        return self.response_position

    def get_maximum_terms(self) -> int:
        if self.config.allow_override_maximum_terms and self.get_extra_request_data(
            SRUParam.X_UNLIMITED_TERMLIST
        ):
            return -1
        if self.maximum_terms == -1:
            return self.config.number_of_terms
        if self.maximum_records > self.config.maximum_terms:
            return self.config.maximum_terms
        return self.maximum_terms

    def get_stylesheet(self) -> Optional[str]:
        return self.stylesheet

    def get_renderBy(self) -> Optional[SRURenderBy]:
        return self.renderBy

    def get_response_type(self) -> Optional[str]:
        return self.response_type

    # ----------------------------------------------------

    # raw/parameter grabby stuff

    def get_version_raw(self) -> Optional[SRUVersion]:
        return self.version

    def get_record_xml_escaping_raw(self) -> Optional[str]:
        if self.is_version(SRUVersion.VERSION_2_0):
            return self.get_parameter(SRUParam.RECORD_XML_ESCAPING, True, False)
        else:
            return self.get_parameter(SRUParam.RECORD_PACKING, True, False)

    def get_record_packing_raw(self) -> Optional[str]:
        if self.is_version(SRUVersion.VERSION_2_0):
            return self.get_parameter(SRUParam.RECORD_PACKING, True, False)
        else:
            return None

    def get_record_schema_identifier_raw(self) -> Optional[str]:
        return self.get_parameter(SRUParam.RECORD_SCHEMA, True, False)

    def get_query_raw(self) -> Optional[str]:
        return self.get_parameter(SRUParam.QUERY, True, False)

    def get_maximum_records_raw(self) -> int:
        return self.maximum_records

    def get_scan_clause_raw(self) -> Optional[str]:
        return self.get_parameter(SRUParam.SCAN_CLAUSE, True, False)

    def get_http_accept_raw(self) -> Optional[str]:
        return self.get_parameter(SRUParam.HTTP_ACCEPT, True, False)

    # FIXME: access request

    def get_indent_response(self) -> int:
        if self.config.allow_override_indent_response:
            value_str = self.get_extra_request_data(SRUParam.X_INDENT_RESPONSE)
            if value_str:
                try:
                    value = int(value_str)
                    if value > -2 and value < 9:
                        return value
                except Exception:
                    pass

        return self.config.indent_response

    def get_http_accept(self) -> Optional[str]:
        if self.http_accept is not None:
            return self.http_accept
        return self.request.headers.get("ACCEPT")

    def get_protocol_schema(self) -> str:
        return "https://" if self.request.is_secure else "http://"

    # ----------------------------------------------------

    def add_diagnostic(
        self, uri: str, details: Optional[str] = None, message: Optional[str] = None
    ) -> None:
        self.add_diagnostic_obj(SRUDiagnostic(uri, details, message))

    def add_diagnostic_obj(self, diagnostic: SRUDiagnostic):
        if self.diagnostics is None:
            self.diagnostics = list()
        self.diagnostics.append(diagnostic)

    # ----------------------------------------------------

    def _parse_number_parameter(self, param: str, value: str, min: int) -> int:
        result = -1

        if value:
            try:
                result = int(value)
                if result < min:
                    self.add_diagnostic(
                        SRUDiagnostics.UNSUPPORTED_PARAMETER_VALUE,
                        param,
                        f"Value is less than {min}.",
                    )
            except Exception:
                self.add_diagnostic(
                    SRUDiagnostics.UNSUPPORTED_PARAMETER_VALUE,
                    param,
                    "Invalid number format.",
                )

        return result

    def _parse_scan_query_parameter(
        self, param: str, value: str
    ) -> Optional[cql.CQLQuery]:
        # NOTE: this should only be called in `check_parameters_rest`
        # when version is not None anymore
        sru_query = CQLQueryParser().parse_query(
            self.version, {SRUParam.QUERY: value}, self  # type: ignore
        )
        if sru_query is None:
            return None
        return sru_query.parsed_query

    def _parse_and_check_version_parameter(
        self, operation: SRUOperation
    ) -> Optional[SRUVersion]:
        version_str = self.get_parameter(SRUParam.VERSION, True, True)
        if version_str:
            if version_str == SRUVersion.VERSION_1_1:
                return SRUVersion.VERSION_1_1
            if version_str == SRUVersion.VERSION_1_2:
                return SRUVersion.VERSION_1_2
            self.add_diagnostic(
                SRUDiagnostics.UNSUPPORTED_VERSION,
                SRUVersion.VERSION_1_2,
                f"Version '{version_str}' is not supported",
            )
            return None

        # except for "explain" operation, complain if "version" parameter
        # was not supplied.
        if operation != SRUOperation.EXPLAIN:
            self.add_diagnostic(
                SRUDiagnostics.MANDATORY_PARAMETER_NOT_SUPPLIED,
                str(SRUParam.VERSION),
                f"Mandatory parameter '{SRUParam.VERSION!s}' was not supplied.",
            )

        # this is an explain operation, assume default version
        return self.config.default_version

    # ----------------------------------------------------

    def check_parameters(self) -> bool:
        """Validate incoming request parameters

        Returns:
            bool: ``True`` if successful, ``False``  if something
                went wrong
        """
        if not self.check_parameters_version_operation():
            return False

        self._check_parameters_rest()
        self._check_parameters_auth()

        # diagnostics is None -> consider as success
        # FIXME: this should be done nicer!
        return not self.diagnostics

    def check_parameters_version_operation(self) -> bool:
        """Validate incoming request parameters **version** and
        **operation**.

        Returns:
            bool: ``True`` if successful, ``False``  if something
                went wrong
        """

        # generally assume, we will also allow processing of SRU 1.1 or 1.2
        process_SRU_old = True

        # Heuristic to detect SRU version and operation ...
        if self.config.max_version >= SRUVersion.VERSION_2_0:
            if not self.get_parameter(SRUParam.VERSION, False, False):
                # Ok, we're committed to SRU 2.0 now, so don't allow processing
                # of SRU 1.1 and 1.2 ...
                process_SRU_old = False

                LOGGER.debug(
                    "handling request as SRU 2.0, because no '%s' parameter was found in the request",
                    SRUParam.VERSION,
                )
                if self.get_parameter(
                    SRUParam.QUERY, False, False
                ) or self.get_parameter(SRUParam.QUERY_TYPE, False, False):
                    LOGGER.debug(
                        "found parameter '%s' or '%s' therefore assuming '%s' operation",
                        SRUParam.QUERY,
                        SRUParam.QUERY_TYPE,
                        SRUOperation.SEARCH_RETRIEVE,
                    )
                    operation = SRUOperation.SEARCH_RETRIEVE
                elif self.get_parameter(SRUParam.SCAN_CLAUSE, False, False):
                    LOGGER.debug(
                        "found parameter '%s' therefore assuming '%s' operation",
                        SRUParam.SCAN_CLAUSE,
                        SRUOperation.SCAN,
                    )
                    operation = SRUOperation.SCAN
                else:
                    LOGGER.debug(
                        "no special parameter found therefore assuming '%s' operation",
                        SRUOperation.EXPLAIN,
                    )
                    operation = SRUOperation.EXPLAIN

                # record version ...
                version: Optional[SRUVersion] = SRUVersion.VERSION_2_0

                # do pedantic check for 'operation' parameter
                operation_str = self.get_parameter(SRUParam.OPERATION, False, False)
                if operation_str:
                    # XXX: if operation is searchRetrive and the 'operation'
                    # parameter is also searchRetrieve, should the server just
                    # ignore it?
                    if (
                        operation != SRUOperation.SEARCH_RETRIEVE
                        and operation_str == SRUOperation.SEARCH_RETRIEVE
                    ):
                        self.add_diagnostic(
                            SRUDiagnostics.UNSUPPORTED_PARAMETER,
                            SRUParam.OPERATION,
                            message=f"Parameter '{SRUParam.OPERATION}' is not valid for SRU version 2.0",
                        )

            else:
                LOGGER.debug(
                    "handling request as legacy SRU, because found parameter '%s' in request",
                    SRUParam.VERSION,
                )

        if process_SRU_old:
            # parse mandatory operation parameter
            operation_str = self.get_parameter(SRUParam.OPERATION, False, False)
            if operation_str:
                if not operation_str.isspace():
                    if operation_str == SRUOperation.EXPLAIN:
                        operation = SRUOperation.EXPLAIN
                    elif operation_str == SRUOperation.SCAN:
                        operation = SRUOperation.SCAN
                    elif operation_str == SRUOperation.SEARCH_RETRIEVE:
                        operation = SRUOperation.SEARCH_RETRIEVE
                    else:
                        self.add_diagnostic(
                            SRUDiagnostics.UNSUPPORTED_OPERATION,
                            message=f"Operation '{operation_str}' is not supported",
                        )
                else:
                    self.add_diagnostic(
                        SRUDiagnostics.UNSUPPORTED_OPERATION,
                        message=f"An empty parameter '{SRUParam.OPERATION}' is not supported.",
                    )

                # parse and check version
                version = self._parse_and_check_version_parameter(operation)
            else:
                # absent parameter should be interpreted as "explain"
                operation = SRUOperation.EXPLAIN
                # parse and check version
                version = self._parse_and_check_version_parameter(operation)

        # sanity check
        if version and operation:
            LOGGER.debug(
                "min = %s, min? = %s, max = %s, max? = %s, version = %s",
                self.config.min_version,
                version == self.config.min_version,
                self.config.max_version,
                version == self.config.max_version,
                version,
            )
            if (
                version >= self.config.min_version
                and version <= self.config.max_version
            ):
                self.version = version
                self.operation = operation

                return True

            else:
                self.add_diagnostic(
                    SRUDiagnostics.UNSUPPORTED_VERSION,
                    self.config.max_version,
                    message=f"Version '{version}' is not supported by this endpoint.",
                )

        LOGGER.debug("bailed")
        return False

    def _check_parameters_rest(self) -> bool:
        """Validate incoming request parameters.

        Returns:
            bool: ``True`` if successful, ``False``  if something
                went wrong
        """
        if self.diagnostics:
            # this should only happen if repeatedly called
            # which is not done usually
            return False

        # check mandatory/optional parameters for operation
        parameters = ParameterInfoSets.for_operation(self.operation)
        if not parameters:
            self.add_diagnostic(
                SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                message="internal error (invalid operation)",
            )
            return False

        # keep list of all submitted parameters (except "operation" and
        # "version"), so we can later warn if an unsupported parameter
        # was sent (= not all parameters were consumed).
        parameter_names = self.get_parameter_names()

        # check parameters ...
        for parameter in parameters:
            name = parameter.name(self.version)  # type: ignore
            if not name:
                # this parameter is not supported in the SRU version that
                # was used for the request
                continue

            value = self.get_parameter(name, True, True)
            if value is None:
                if parameter.mandatory:
                    self.add_diagnostic(
                        SRUDiagnostics.MANDATORY_PARAMETER_NOT_SUPPLIED,
                        name,
                        message=f"Mandatory parameter '{name}' was not supplied.",
                    )
                continue

            # remove supported parameter from list
            if name in parameter_names:
                parameter_names.remove(name)

            # if parameter is not supported in this version, skip it
            # and create add an diagnostic.
            # NOTE: version is not None
            if not parameter.is_for_version(self.version):  # type: ignore
                self.add_diagnostic(
                    SRUDiagnostics.UNSUPPORTED_PARAMETER,
                    name,
                    message=f"Version {self.version} does not support parameter '{name}'.",
                )
                continue

            # validate and parse parameters ...
            if parameter.parameter == ParameterInfo.Parameter.RECORD_XML_ESCAPING:
                if value == SRUParamValue.RECORD_XML_ESCAPING_XML:
                    self.record_xml_escaping = SRURecordXmlEscaping.XML
                elif value == SRUParamValue.RECORD_XML_ESCAPING_STRING:
                    self.record_xml_escaping = SRURecordXmlEscaping.STRING
                else:
                    self.add_diagnostic(
                        SRUDiagnostics.UNSUPPORTED_XML_ESCAPING_VALUE,
                        message=f"Record XML escaping '{value}' is not supported.",
                    )

            elif parameter.parameter == ParameterInfo.Parameter.RECORD_PACKING:
                if value == SRUParamValue.RECORD_PACKING_PACKED:
                    self.record_packing = SRURecordPacking.PACKED
                elif value == SRUParamValue.RECORD_PACKING_UNPACKED:
                    self.record_packing = SRURecordPacking.UNPACKED
                else:
                    self.add_diagnostic(
                        SRUDiagnostics.UNSUPPORTED_PARAMETER_VALUE,
                        message=f"Record packing '{value}' is not supported.",
                    )

            elif parameter.parameter == ParameterInfo.Parameter.RENDER_BY:
                if value == SRUParamValue.RENDER_BY_CLIENT:
                    self.renderBy = SRURenderBy.CLIENT
                elif value == SRUParamValue.RENDER_BY_SERVER:
                    self.renderBy = SRURenderBy.SERVER
                else:
                    self.add_diagnostic(
                        SRUDiagnostics.UNSUPPORTED_PARAMETER_VALUE,
                        message=f"Value '{value}' for parameter '{name}' is not supported.",
                    )

            elif parameter.parameter == ParameterInfo.Parameter.RECORD_SCHEMA:
                # The parameter recordSchema may contain either schema
                # identifier or the short name. If available, set to
                # appropriate schema identifier in the request object.
                schema_info = self.config.find_schema_info(value)
                if schema_info:
                    self.record_schema_identifier = schema_info.identifier
                else:
                    # SRU servers are supposed to raise a non-surrogate
                    # (fatal) diagnostic in case the record schema is not
                    # known to the server.
                    self.add_diagnostic(
                        SRUDiagnostics.UNKNOWN_SCHEMA_FOR_RETRIEVAL,
                        value,
                        message=f"Record schema '{value}' is not supported for retrieval.",
                    )

            elif parameter.parameter == ParameterInfo.Parameter.START_RECORD:
                self.start_record = self._parse_number_parameter(name, value, 1)
            elif parameter.parameter == ParameterInfo.Parameter.RESPONSE_POSITION:
                self.response_position = self._parse_number_parameter(name, value, 0)
            elif parameter.parameter == ParameterInfo.Parameter.MAXIMUM_RECORDS:
                self.maximum_records = self._parse_number_parameter(name, value, 0)
            elif parameter.parameter == ParameterInfo.Parameter.MAXIMUM_TERMS:
                self.maximum_terms = self._parse_number_parameter(name, value, 0)
            elif parameter.parameter == ParameterInfo.Parameter.RESULT_SET_TTL:
                self.resultSet_TTL = self._parse_number_parameter(name, value, 0)

            elif parameter.parameter == ParameterInfo.Parameter.SCAN_CLAUSE:
                self.scan_clause = self._parse_scan_query_parameter(name, value)

            elif parameter.parameter == ParameterInfo.Parameter.RECORD_XPATH:
                self.record_xpath = value
            elif parameter.parameter == ParameterInfo.Parameter.SORT_KEYS:
                self.sortKeys = value
            elif parameter.parameter == ParameterInfo.Parameter.STYLESHEET:
                self.stylesheet = value

            elif parameter.parameter == ParameterInfo.Parameter.RESPONSE_TYPE:
                # FIXME: check parameter validity?!
                self.response_type = value
            elif parameter.parameter == ParameterInfo.Parameter.HTTP_ACCEPT:
                # FIXME: check parameter validity?!
                self.http_accept = value

        # handle query and queryType
        if self.operation == SRUOperation.SEARCH_RETRIEVE:
            # determine queryType
            query_type: Optional[str] = None
            if self.version == SRUVersion.VERSION_2_0:
                if SRUParam.QUERY_TYPE in parameter_names:
                    parameter_names.remove(SRUParam.QUERY_TYPE)
                value = self.get_parameter(SRUParam.QUERY_TYPE, True, True)
                if value is None:
                    query_type = SRUQueryType.CQL.value
                else:
                    has_bad_chars = QUERY_TYPE_ALLOWED_CHARS.fullmatch(value) is None
                    if has_bad_chars:
                        self.add_diagnostic(
                            SRUDiagnostics.UNSUPPORTED_PARAMETER_VALUE,
                            SRUParam.QUERY_TYPE,
                            message="Value contains illegal characters.",
                        )
                    else:
                        query_type = value
            else:
                # SRU 1.1 and SRU 1.2 only support CQL
                query_type = SRUQueryType.CQL.value

            if query_type:
                LOGGER.debug("looking for query parser for query type '%s'", query_type)
                query_parser = self.query_parsers.find_query_parser(query_type)
                if query_parser:
                    if query_parser.supports_version(self.version):
                        # gather query parameters
                        # (as required by QueryParser implementation)
                        query_parameters = dict()
                        missing_parameters = list()
                        for name in query_parser.query_parameter_names:
                            if name in parameter_names:
                                parameter_names.remove(name)
                            value = self.get_parameter(name, True, False)
                            if value is not None:
                                query_parameters[name] = value
                            else:
                                missing_parameters.append(name)

                        if not missing_parameters:
                            LOGGER.debug(
                                "parsing query with parser for type '%s' and parameters %s",
                                query_parser.query_type,
                                query_parameters,
                            )
                            # NOTE: version is not None
                            self.query = query_parser.parse_query(
                                self.version, query_parameters, self  # type: ignore
                            )
                            if not self.query:
                                LOGGER.debug("query parser failed to parse query")
                                self.add_diagnostic(
                                    SRUDiagnostics.QUERY_SYNTAX_ERROR,
                                    message="Query could not be parsed.",
                                )
                        else:
                            LOGGER.debug(
                                "parameters %s missing, cannot parse query",
                                missing_parameters,
                            )
                            for name in missing_parameters:
                                self.add_diagnostic(
                                    SRUDiagnostics.MANDATORY_PARAMETER_NOT_SUPPLIED,
                                    name,
                                    message=(
                                        f"Mandatory parameter '{name}' is missing or empty. "
                                        f"Required to perform query of query type '{query_type}'."
                                    ),
                                )

                    else:
                        LOGGER.debug(
                            "query parser for query type '%s' is not supported by SRU version %s",
                            query_type,
                            self.version,
                        )
                        self.add_diagnostic(
                            SRUDiagnostics.CANNOT_PROCESS_QUERY_REASON_UNKNOWN,
                            message=(
                                f"Query parser for query type '{query_type}' is not"
                                f" supported by SRU version '{self.version}'."
                            ),
                        )

                else:
                    LOGGER.debug("no parser for query type '%s' found", query_type)
                    self.add_diagnostic(
                        SRUDiagnostics.CANNOT_PROCESS_QUERY_REASON_UNKNOWN,
                        message=f"Cannot find query parser for query type '{query_type}'.",
                    )

            else:
                LOGGER.debug("cannot determine query type")
                self.add_diagnostic(
                    SRUDiagnostics.CANNOT_PROCESS_QUERY_REASON_UNKNOWN,
                    message="Cannot determine query type.",
                )

        # check if any parameters where not consumed and
        # add appropriate warnings
        if parameter_names:
            for name in parameter_names:
                # skip extraRequestData (aka extensions)
                if not name.startswith(PARAM_EXTENSION_PREFIX):
                    self.add_diagnostic(
                        SRUDiagnostics.UNSUPPORTED_PARAMETER,
                        name,
                        message=f"Parameter '{name}' is not supported for this operation.",
                    )

        # diagnostics is None -> consider as success
        # FIXME: this should be done nicer!
        return not self.diagnostics

    def _check_parameters_auth(self) -> None:
        # extract authentication information from,
        # if an authentication provider is set
        if self.authentication_info_provider:
            try:
                self.authentication_info = (
                    self.authentication_info_provider.get_AuthenticationInfo(
                        self.request
                    )
                )
            except SRUException as ex:
                self.add_diagnostic_obj(ex.get_diagnostic())

    # ----------------------------------------------------

    def get_parameter_names(self) -> List[str]:
        parameters = list(self.request.args.keys())
        parameters = [
            p for p in parameters if p not in (SRUParam.OPERATION, SRUParam.VERSION)
        ]
        return parameters

    def get_parameter(
        self, name: Union[SRUParam, str], nullify: bool, diagnostic_if_empty: bool
    ) -> Optional[str]:
        value = self.request.args.get(name)
        if value is not None:
            value = value.strip()
            if nullify and not value:
                value = None
                if diagnostic_if_empty:
                    self.add_diagnostic(
                        SRUDiagnostics.UNSUPPORTED_PARAMETER_VALUE,
                        name,
                        message=f"An empty parameter '{name}' is not supported.",
                    )
        return value

    def get_extra_request_data_names(self) -> List[str]:
        parameters = list(self.request.args.keys())
        parameters = [p for p in parameters if p.startswith(PARAM_EXTENSION_PREFIX)]
        return parameters

    def get_extra_request_data(self, name: str) -> Optional[str]:
        if name is None:
            raise TypeError("name is None")
        if not name.startswith(PARAM_EXTENSION_PREFIX):
            raise ValueError(f"name must start with '{PARAM_EXTENSION_PREFIX}'")
        return self.request.args.get(name)


# ---------------------------------------------------------------------------



import pytest

from clarin.sru.constants import SRUDiagnostics
from clarin.sru.diagnostic import SRUDiagnostic
from clarin.sru.exception import SRUException

# ---------------------------------------------------------------------------


def test_SRUDiagnostic():
    with pytest.raises(TypeError):
        diag = SRUDiagnostic()

    diag = SRUDiagnostic("my-uri")
    assert diag.uri == "my-uri"
    assert diag.message is None
    assert diag.details is None

    # best to supply the URI via Enum
    diag = SRUDiagnostic(SRUDiagnostics.GENERAL_SYSTEM_ERROR)
    assert diag.uri == SRUDiagnostics.GENERAL_SYSTEM_ERROR.value
    assert isinstance(diag.uri, str)
    assert SRUDiagnostics in diag.uri.__class__.mro()
    assert str in diag.uri.__class__.mro()
    # message is set by Enum description
    assert diag.message == SRUDiagnostics.GENERAL_SYSTEM_ERROR.description
    assert diag.details is None

    # but can supply just the URI as string
    diag = SRUDiagnostic(SRUDiagnostics.GENERAL_SYSTEM_ERROR.value)
    assert diag.uri == SRUDiagnostics.GENERAL_SYSTEM_ERROR.value
    assert isinstance(diag.uri, str)
    assert SRUDiagnostics not in diag.uri.__class__.mro()
    assert str in diag.uri.__class__.mro()
    assert diag.message == SRUDiagnostics.GENERAL_SYSTEM_ERROR.description
    assert diag.details is None

    # overwrite message (don't use default)
    diag = SRUDiagnostic(SRUDiagnostics.GENERAL_SYSTEM_ERROR, message="test")
    assert diag.uri == SRUDiagnostics.GENERAL_SYSTEM_ERROR.value
    assert diag.message == "test"
    assert diag.details is None

    diag = SRUDiagnostic(
        SRUDiagnostics.GENERAL_SYSTEM_ERROR, details="abc", message="test"
    )
    assert diag.uri == SRUDiagnostics.GENERAL_SYSTEM_ERROR.value
    assert diag.message == "test"
    assert diag.details == "abc"

    diag = SRUDiagnostic(SRUDiagnostics.GENERAL_SYSTEM_ERROR, details="abc")
    assert diag.uri == SRUDiagnostics.GENERAL_SYSTEM_ERROR.value
    assert diag.message == SRUDiagnostics.GENERAL_SYSTEM_ERROR.description
    assert diag.details == "abc"


def test_diagnostic_from_error():
    exc = SRUException(
        SRUDiagnostics.AUTHENTICATION_ERROR, details="abc", message="def"
    )
    diag = exc.get_diagnostic()
    assert diag.uri == SRUDiagnostics.AUTHENTICATION_ERROR
    assert diag.details == "abc"
    assert diag.message == "def"

    exc = SRUException(SRUDiagnostics.AUTHENTICATION_ERROR)
    diag = exc.get_diagnostic()
    assert diag.uri == SRUDiagnostics.AUTHENTICATION_ERROR
    assert diag.details is None
    assert diag.message == SRUDiagnostics.AUTHENTICATION_ERROR.description


# ---------------------------------------------------------------------------



from typing import Dict
from typing import Optional
from typing import Union

import pytest

from clarin.sru.diagnostic import SRUDiagnosticList
from clarin.sru.server.config import SRUServerConfig
from clarin.sru.server.config import SRUServerConfigKey
from clarin.sru.server.request import SRURequest
from clarin.sru.server.result import SRUExplainResult
from clarin.sru.server.result import SRUScanResultSet
from clarin.sru.server.result import SRUSearchResultSet
from clarin.sru.xml.writer import SRUXMLStreamWriter

ConfigParamDict = Dict[Union[SRUServerConfigKey, str], str]


# ---------------------------------------------------------------------------


@pytest.fixture(scope="function")
def config_params() -> ConfigParamDict:
    # minimal required params
    return {
        SRUServerConfigKey.SRU_TRANSPORT: "http",
        SRUServerConfigKey.SRU_HOST: "localhost",
        SRUServerConfigKey.SRU_PORT: "80",
        SRUServerConfigKey.SRU_DATABASE: "test",
        SRUServerConfigKey.SRU_ALLOW_OVERRIDE_INDENT_RESPONSE: "true",
    }


# ---------------------------------------------------------------------------


def test_server(server_config_file, config_params):
    import logging

    from werkzeug.test import Client
    from werkzeug.testapp import test_app

    from clarin.sru.queryparser import SRUQueryParserRegistry
    from clarin.sru.server.server import SRUServer

    logging.basicConfig(level=logging.DEBUG)

    class FakeSE:
        def explain(self, *args, **kwargs):
            return None

    config = SRUServerConfig.parse(config_params, server_config_file)
    query_parsers = SRUQueryParserRegistry.Builder(True).build()
    server = SRUServer(config, query_parsers, FakeSE())

    client = Client(test_app)
    response = client.get(
        "?operation=explain&x-fcs-endpoint-description=true&x-indent-response=1"
    )
    request = response.request

    server.handle_request(request, response)

    print(response.get_data().decode())

    assert len(response.get_data()) == 1776


def test_server_app_check(server_config_file, config_params):
    from clarin.sru.exception import SRUConfigException
    from clarin.sru.server.wsgi import SRUServerApp

    class FakeSE:
        def explain(self, *args, **kwargs):
            return None

    err_msg = "Error creating search engine"
    err_msg2 = "Parameter 'SRUSearchEngine_clazz' not of type SRUSearchEngine"
    with pytest.raises(
        RuntimeError, match=f"error configuring or inializing the server: {err_msg}"
    ) as exc_info:
        SRUServerApp(FakeSE, server_config_file, config_params, develop=False)
    assert isinstance(exc_info.value.__cause__, SRUConfigException)
    assert exc_info.value.__cause__.args[0] == err_msg
    assert isinstance(exc_info.value.__cause__.__cause__, ValueError)
    assert exc_info.value.__cause__.__cause__.args[0] == err_msg2


def test_server_app_explain(server_config_file, config_params):
    import logging

    from werkzeug.test import Client

    from clarin.sru.server.server import SRUSearchEngine
    from clarin.sru.server.wsgi import SRUServerApp

    logging.basicConfig(level=logging.DEBUG)

    class TestSE(SRUSearchEngine):
        def explain(
            self,
            config: SRUServerConfig,
            request: SRURequest,
            diagnostics: SRUDiagnosticList,
        ) -> Optional[SRUExplainResult]:
            return None

        def scan(
            self,
            config: SRUServerConfig,
            request: SRURequest,
            diagnostics: SRUDiagnosticList,
        ) -> Optional[SRUScanResultSet]:
            return None

        def search(
            self,
            config: SRUServerConfig,
            request: SRURequest,
            diagnostics: SRUDiagnosticList,
        ) -> SRUSearchResultSet:
            return super().search(config, request, diagnostics)

    app = SRUServerApp(TestSE, server_config_file, config_params, develop=False)

    client = Client(app)

    response = client.get()
    print(response.get_data().decode())
    assert len(response.get_data()) == 1726

    response = client.get("?operation=explain&x-fcs-endpoint-description=true")
    print(response.get_data().decode())
    assert len(response.get_data()) == 1726

    response = client.get(
        "?operation=explain&x-fcs-endpoint-description=true&x-indent-response=1"
    )
    print(response.get_data().decode())
    assert len(response.get_data()) == 1943


def test_server_app_search(server_config_file, config_params):
    import logging

    from werkzeug.test import Client

    from clarin.sru.server.server import SRUSearchEngine
    from clarin.sru.server.wsgi import SRUServerApp

    logging.basicConfig(level=logging.DEBUG)

    class TestSRS(SRUSearchResultSet):
        def __init__(self, diagnostics: SRUDiagnosticList) -> None:
            super().__init__(diagnostics)
            self.is_first_result = True

        def get_record_count(self) -> int:
            return 1

        def get_record_identifier(self) -> str:
            return "t1"

        def get_record_schema_identifier(self) -> str:
            return "test"

        def get_total_record_count(self) -> int:
            return 1

        def next_record(self) -> bool:
            if self.is_first_result:
                self.is_first_result = False
                return True
            return False

        def write_record(self, writer: SRUXMLStreamWriter) -> None:
            writer.characters("test1")

    class TestSE(SRUSearchEngine):
        def explain(
            self,
            config: SRUServerConfig,
            request: SRURequest,
            diagnostics: SRUDiagnosticList,
        ) -> Optional[SRUExplainResult]:
            return None

        def scan(
            self,
            config: SRUServerConfig,
            request: SRURequest,
            diagnostics: SRUDiagnosticList,
        ) -> Optional[SRUScanResultSet]:
            return None

        def search(
            self,
            config: SRUServerConfig,
            request: SRURequest,
            diagnostics: SRUDiagnosticList,
        ) -> SRUSearchResultSet:
            return TestSRS(diagnostics)

    app = SRUServerApp(TestSE, server_config_file, config_params, develop=False)

    client = Client(app)

    response = client.get()
    print(response.get_data().decode())
    assert len(response.get_data()) == 1726

    response = client.get("?operation=searchRetrieve&query=dog&x-indent-response=1")
    data = response.get_data()
    print(data.decode())
    assert len(data) == 1101
    assert b"<sruResponse:recordData>test1</sruResponse:recordData>" in data
    assert b"<sruResponse:numberOfRecords>1</sruResponse:numberOfRecords>" in data



from abc import ABC
from abc import ABCMeta
from abc import abstractmethod
from enum import Enum
from typing import Optional

from ..constants import SRUResultCountPrecision
from ..diagnostic import SRUDiagnostic
from ..diagnostic import SRUDiagnosticList
from ..xml.writer import SRUXMLStreamWriter

# ---------------------------------------------------------------------------


class SRUAbstractResult(metaclass=ABCMeta):
    """Base class for SRU responses."""

    def __init__(self, diagnostics: SRUDiagnosticList) -> None:
        if diagnostics is None:
            raise TypeError("Implementation error: diagnostics must not be None!")
        self.diagnostics = diagnostics

    def add_diagnostic(
        self, uri: str, details: Optional[str] = None, message: Optional[str] = None
    ) -> None:
        """Add a non surrogate diagnostic to the response.

        Args:
            uri: the diagnostic's identifying URI
            details: supplementary information available, often
                in a format specified by the diagnostic or ``None``
            message: human readable message to display to the
                end user or ``None``
        """
        self.diagnostics.add_diagnostic(uri, details, message)

    @property
    def has_extra_response_data(self) -> bool:
        """Check, if extra response data should be serialized for
        this request. Default implementation is provided for convince
        and always returns ``False``.

        Returns:
            bool: ``True`` if extra response data should be serialized
        """
        return False

    def write_extra_response_data(self, writer: SRUXMLStreamWriter) -> None:
        """Serialize extra response data for this request. A no-op
        default implementation is provided for convince.

        Args:
            writer: Writer to serialize extra response data
        """

    def close(self) -> None:
        """Release this result and free any associated resources.

        This method **must not** throw any exceptions.

        Calling the method `close` on a result object that
        is already closed is a no-op.
        """


# ---------------------------------------------------------------------------


class SRUExplainResult(ABC, SRUAbstractResult):
    """A result set of an ``explain`` operation. A database
    implementation may use it implement extensions to the SRU
    protocol, i.e. providing extraResponseData.

    This class needs to be implemented for the target data source.

    See also:
        SRU Explain Operation: http://www.loc.gov/standards/sru/explain/
    """


# ---------------------------------------------------------------------------


class SRUScanResultSet(ABC, SRUAbstractResult):
    """A result set of a ``scan`` operation. It is used to iterate
    over the term set and provides a method to serialize the terms.

    A `SRUScanResultSet` object maintains a cursor pointing to its
    current term. Initially the cursor is positioned before the first
    term. The `next` method moves the cursor to the next term, and
    because it returns ``False`` when there are no more terms in the
    `SRUScanResultSet` object, it can be used in a `while` loop to
    iterate through the term set.

    This class needs to be implemented for the target search engine.

    See also:
        SRU Scan Operation: http://www.loc.gov/standards/sru/companionSpecs/scan.html
    """

    class WhereInList(str, Enum):
        """A flag to indicate the position of the term within the
        complete term list."""

        FIRST = "first"
        """The first term (**first**)"""

        LAST = "last"
        """The last term (**last**)"""

        ONLY = "only"
        """The only term (**only**)"""

        INNER = "inner"
        """Any other term (**inner**)"""

    @abstractmethod
    def next_term(self) -> bool:
        """Moves the cursor forward one term from its current
        position. A result set cursor is initially positioned before
        the first record; the first call to the method `next` makes
        the first term the current term; the second call makes the
        second term the current term, and so on.

        When a call to the `next` method returns ``False``, the
        cursor is positioned after the last term.

        Returns:
            bool: ``True`` if the new current term is valid;
                ``False`` if there are no more terms

        Raises:
            `SRUException`: if an error occurred while fetching the
                next term
        """

    @abstractmethod
    def get_value(self) -> str:
        """Get the current term exactly as it appears in the index.

        Returns:
            str: current term
        """

    @abstractmethod
    def get_number_of_records(self) -> int:
        """Get the number of records for the current term which would
        be matched if the index in the request's `scanClause` was
        searched with the term in the `value` field.

        Returns:
            int: a non-negative number of records or ``-1``, if the
                number is unknown.
        """

    @abstractmethod
    def get_display_term(self) -> Optional[str]:
        """Get the string for the current term to display to the end
        user in place of the term itself.

        Returns:
            str: display string or ``None``
        """

    @abstractmethod
    def get_WhereInList(self) -> Optional[WhereInList]:
        """Get the flag to indicate the position of the term within
        the complete term list.

        Returns:
            `WhereInList`: position within term list or ``None``
        """

    def has_extra_term_data(self) -> bool:
        """Check, if extra term data should be serialized for the
        current term. A default implementation is provided for
        convince and always returns ``False``.

        Returns:
            bool: ``True`` if the term has extra term data

        Raises:
            `StopIteration`: term set is already advanced past all
                past terms

        See also:
            `write_extra_term_data`
        """
        return False

    @abstractmethod
    def write_extra_term_data(self, writer: SRUXMLStreamWriter):
        """Serialize extra term data for the current term. A no-op
        default implementation is provided for convince.

        Args:
            writer: Writer to serialize extra term data for current
                term

        Raises:
            `StopIteration`: term set already advanced past all terms
        """


# ---------------------------------------------------------------------------


class SRUSearchResultSet(ABC, SRUAbstractResult):
    """A result set of a ``searchRetrieve`` operation. It it used to
    iterate over the result set and provides a method to serialize
    the record in the requested format.

    A `SRUSearchResultSet` object maintains a cursor pointing to its
    current record. Initially the cursor is positioned before the
    first record. The `next` method moves the cursor to the next
    record, and because it returns ``False`` when there are no more
    records in the `SRUSearchResultSet` object, it can be used in a
    `while` loop to iterate through the result set.

    This class needs to be implemented for the target search engine.

    See also:
        * SRU Search Retrieve Operation: http://www.loc.gov/standards/sru/
        * SRU 1.1 SR: http://www.loc.gov/standards/sru/sru-1-1.html
        * SRU 1.2 SR: http://www.loc.gov/standards/sru/sru-1-2.html
        * SRU 2.0 SR: http://www.loc.gov/standards/sru/sru-2-0.html
        * Differences SRU 2.0 to SRU 1.2: http://www.loc.gov/standards/sru/differences.html
    """

    @abstractmethod
    def get_total_record_count(self) -> int:
        """The number of records matched by the query. If the query
        fails this must be ``0``. If the search engine cannot
        determine the total number of matched by a query, it must
        return ``-1``.

        Returns:
            int: the total number of results or ``0`` if the query
                failed or ``-1`` if the search engine cannot
                determine the total number of results
        """

    @abstractmethod
    def get_record_count(self) -> int:
        """The number of records matched by the query but at most as
        the number of records requested to be returned
        (``maximumRecords`` parameter). If the query fails this must
        be ``0``.

        Returns:
            int: the number of results or ``0`` if the query failed
        """

    def get_resultSet_id(self) -> Optional[str]:
        """The result set id of this result. The default
        implementation returns ``None``.

        Returns:
            str: the result set id or ``None`` if not applicable for
                this result
        """
        return None

    def get_resultSet_TTL(self) -> int:
        """The result set time to live. In SRU 2.0 it will be
        serialized as ``<resultSetTTL>`` element; in SRU 1.2 as
        ``<resultSetIdleTime>`` element.The default implementation
        returns ``-1``.

        Returns:
            int: the result set time to live or ``-1`` if not
                applicable for this result
        """
        return -1

    def get_result_count_precision(self) -> Optional[SRUResultCountPrecision]:
        """(SRU 2.0) Indicate the accuracy of the result count
        reported by total number of records that matched the query.
        Default implementation returns ``None``.

        Returns:
            Optional[SRUResultCountPrecision]: the result count
                precision or ``None`` if not applicable for this
                result

        See also:
            `SRUResultCountPrecision`
        """
        return None

    @abstractmethod
    def get_record_schema_identifier(self) -> str:
        """The record schema identifier in which the records are
        returned (``recordSchema`` parameter).

        Returns:
            str: the record schema identifier
        """

    @abstractmethod
    def next_record(self) -> bool:
        """Moves the cursor forward one record from its current
        position. A `SRUSearchResultSet` cursor is initially
        positioned before the first record; the first call to the
        method `next` makes the first record the current record; the
        second call makes the second record the current record, and
        so on.

        When a call to the `next` method returns ``False``, the
        cursor is positioned after the last record.

        Returns:
            bool: ``True`` if the new current record is valid;
                ``False`` if there are no more records

        Raises:
            `SRUException`: if an error occurred while fetching the
                next record
        """

    @abstractmethod
    def get_record_identifier(self) -> Optional[str]:
        """An identifier for the current record by which it can
        unambiguously be retrieved in a subsequent operation.

        Returns:
            str: identifier for the record or ``None`` of none is
                available

        Raises:
            `StopIteration`: result set is past all records
        """

    def get_surrogate_diagnostic(self) -> Optional[SRUDiagnostic]:
        """Get surrogate diagnostic for current record. If this
        method returns a diagnostic, the `write_record method will
        not be called. The default implementation returns ``None``.

        Returns:
            Optional[SRUDiagnostic]: a surrogate diagnostic or
                ``None``
        """
        return None

    @abstractmethod
    def write_record(self, writer: SRUXMLStreamWriter) -> None:
        """Serialize the current record in the requested format.

        Args:
            writer: Writer to serialize current record

        Raises:
            `StopIteration`: result set is past all records
        """

    @property
    def has_extra_record_data(self) -> bool:
        """Check, if extra record data should be serialized for the
        current record. The default implementation returns ``False``.

        Returns:
            bool: ``True`` if the record has extra record data

        Raises:
            `StopIteration`: result set is past all records

        See also:
            `write_extra_record_data`
        """
        return False

    def write_extra_record_data(self, writer: SRUXMLStreamWriter) -> None:
        """Serialize extra record data for the current record. A
        no-op default implementation is provided for convince.

        Args:
            writer: Writer to serialize extra record data for current
                record

        Raises:
            `StopIteration`: result set past already advanced past
                all records
        """
        pass


# ---------------------------------------------------------------------------



from typing import Optional

from .diagnostic import SRUDiagnostic

# ---------------------------------------------------------------------------


class SRUException(Exception):
    """An exception raised, if something went wrong processing the
    request. For diagnostic codes, see constants in `SRUConstant`.

    See also:
        SRUConstant
    """

    def __init__(
        self,
        uri: str,
        details: Optional[str] = None,
        message: Optional[str] = None,
        *args
    ):
        if uri is None:
            raise TypeError("uri is None")
        if not uri.strip():
            raise ValueError("uri is empty")

        super().__init__(message, *args)
        self.uri = uri
        self.details = details

    def __str__(self):
        # check whether no message was supplied
        if not self.args[0]:
            # check whether this exception was chained (raised as wrapper)
            # then use the message of the 'inner' exception
            #   e.g. with raise SRUException() from exception
            if self.__cause__:
                return self.__cause__.args[0]
            # otherwise get diagnostics message from `uri`
            return SRUDiagnostic.get_default_error_message(self.uri)
        # message was supplied to constructor
        return super().__str__()

    def get_diagnostic(self) -> SRUDiagnostic:
        """Create a SRU diagnostic from this exception."""
        return SRUDiagnostic(self.uri, self.details, self.args[0])


# ---------------------------------------------------------------------------


class SRUConfigException(Exception):
    """An exception raised, if some error occurred with the SRUServer
    configuration."""


# ---------------------------------------------------------------------------



import io
from xml.sax.saxutils import XMLGenerator

import cql
import pytest
from lxml import etree

from clarin.sru.constants import SRURecordXmlEscaping
from clarin.sru.xml.writer import SRUXMLStreamWriter

# ---------------------------------------------------------------------------


def test_streamwriter():
    buf = io.StringIO()

    writer = SRUXMLStreamWriter(buf, SRURecordXmlEscaping.STRING, indent=2)
    # writer = XMLGenerator(buf, "utf-8")

    writer.startDocument()
    writer.startElement("root", dict())
    writer.startElement("ele1", dict())
    writer.endElement("ele1")
    writer.endElement("root")
    writer.endDocument()

    content = buf.getvalue()

    assert (
        content
        == """<?xml version="1.0" encoding="utf-8"?>\n<root>\n  <ele1></ele1>\n</root>"""
    )


def test_streamwriter_xcql():
    buf = io.StringIO()

    writer = SRUXMLStreamWriter(buf, SRURecordXmlEscaping.STRING, indent=2)

    query = cql.parse("stuff")
    query.setServerDefaults()

    writer.startDocument()
    writer.startElement("root", dict())
    writer.startElement("ele1", dict())
    writer.endElement("ele1")
    writer.startElement("xQuery", dict())
    writer.writeXCQL(query, False)
    writer.endElement("xQuery")
    writer.endElement("root")
    writer.endDocument()

    content = buf.getvalue()

    assert (
        content
        == """<?xml version="1.0" encoding="utf-8"?>\n<root>\n  <ele1></ele1>\n"""
        """  <xQuery>\n    <index>cql.serverChoice</index>\n    <relation>\n"""
        """      <value>=</value>\n    </relation>\n    <term>stuff</term>\n"""
        """  </xQuery>\n</root>"""
    )


def test_streamwriter_ns():
    buf = io.StringIO()

    writer = SRUXMLStreamWriter(buf, SRURecordXmlEscaping.STRING, indent=2)
    # writer = XMLGenerator(buf, "utf-8")

    writer.startDocument()
    writer.startElement("root", dict())
    writer.startPrefixMapping("abc", "http://a.bc")
    writer.startElementNS(("http://a.bc", "ele1"), "ele1", dict())
    # writer.startElement("test", dict())
    # writer.endElement("test")
    # writer.startElementNS(("http://a.bc", "test"), None, dict())
    # writer.endElementNS(("http://a.bc", "test"), "test")
    writer.endElementNS(("http://a.bc", "ele1"), "ele1")
    writer.endElement("root")
    writer.endPrefixMapping("abc")
    writer.endDocument()

    content = buf.getvalue()

    assert (
        content == """<?xml version="1.0" encoding="utf-8"?>\n<root>\n"""
        """  <abc:ele1 xmlns:abc="http://a.bc"></abc:ele1>\n</root>"""
    )


def test_streamwriter_xcql_ns():
    buf = io.StringIO()

    writer = SRUXMLStreamWriter(buf, SRURecordXmlEscaping.STRING, indent=2)

    query = cql.parse("stuff")
    query.setServerDefaults()

    writer.startDocument()
    writer.startPrefixMapping("abc", "http://a.bc")
    writer.startElement("root", dict())
    writer.startElement("ele1", dict())
    writer.endElement("ele1")
    writer.startElementNS(("http://a.bc", "xQuery"), "xQuery", dict())
    writer.writeXCQL(query, False)
    writer.endElementNS(("http://a.bc", "xQuery"), "xQuery")
    writer.endElement("root")
    writer.endPrefixMapping("abc")
    writer.endDocument()

    content = buf.getvalue()

    assert (
        content
        == """<?xml version="1.0" encoding="utf-8"?>\n<root>\n  <ele1></ele1>\n"""
        """  <abc:xQuery xmlns:abc="http://a.bc">\n    <index>cql.serverChoice"""
        """</index>\n    <relation>\n      <value>=</value>\n    </relation>\n"""
        """    <term>stuff</term>\n  </abc:xQuery>\n</root>"""
    )


def test_record_packing():
    buf = io.StringIO()
    writer = SRUXMLStreamWriter(buf, SRURecordXmlEscaping.XML, indent=2)

    writer.startElement("root", dict())
    writer.startRecord()
    writer.startElement("test", dict())
    writer.endElement("test")
    writer.endRecord()
    writer.endElement("root")

    content = buf.getvalue()
    assert content == """<root>\n  <test></test>\n</root>"""

    buf = io.StringIO()
    writer = SRUXMLStreamWriter(buf, SRURecordXmlEscaping.STRING, indent=2)

    writer.startElement("root", dict())
    writer.startRecord()
    writer.startElement("test", dict())
    writer.endElement("test")
    writer.endRecord()
    writer.endElement("root")

    content = buf.getvalue()
    assert content == """<root>&lt;test&gt;&lt;/test&gt;</root>"""


def test_default_ns():
    buf = io.StringIO()
    writer = SRUXMLStreamWriter(buf, SRURecordXmlEscaping.XML, indent=2)

    writer.startPrefixMapping(None, "http://default.ns")
    writer.startPrefixMapping("ns", "http://name.space")
    with writer.element("root", "http://name.space"):
        with writer.element("test"):
            writer.characters("content")

    content = buf.getvalue()
    assert (
        content
        == """<ns:root xmlns="http://default.ns" xmlns:ns="http://name.space">\n  <test>content</test>\n</ns:root>"""
    )


# ---------------------------------------------------------------------------


@pytest.mark.skip("manual testing")
def test_buffered_io():
    fp = io.StringIO()
    buf = io.BufferedWriter(fp, buffer_size=16)

    buf.write("abc")

    assert fp.getvalue() == ""
    buf.flush()
    assert fp.getvalue() == "abc"


@pytest.mark.skip("manual testing")
def test_lxml_xmlfile():
    fp = io.BytesIO()
    buf = io.BufferedWriter(fp, buffer_size=1024)

    with etree.xmlfile(buf, encoding="utf-8", close=False, buffered=True) as xf:
        # xf.write_declaration(standalone=True)
        with xf.element("root"):
            with xf.element("elements"):
                for i in range(10):
                    with xf.element("element", {"nr": str(i)}):
                        pass

        xf.flush()

    assert fp.getvalue() == b""

    buf.flush()

    content = fp.getvalue()

    assert content.startswith(b"<root><")

    import lxml.sax

    class MyHandler(lxml.sax.ElementTreeContentHandler):
        def __init__(self, makeelement=None):
            super().__init__(makeelement)
            self.indentlevel = 0
            self.saw_content = False

        def startElementNS(self, ns_name, qname, attributes=None):
            print(ns_name, qname, attributes)
            if qname != "root":
                self.indentlevel += 1
                self.characters("\n" + (self.indentlevel * "  "))
            return super().startElementNS(ns_name, qname, attributes)

        def endElementNS(self, ns_name, qname):
            if not self.saw_content:
                self.characters("\n" + (self.indentlevel * "  "))
                self.indentlevel -= 1
            return super().endElementNS(ns_name, qname)

    handler = MyHandler()

    lxml.sax.saxify(etree.parse(io.BytesIO(content)), handler)

    tree = handler.etree
    content2 = etree.tostring(tree.getroot())
    print(content2.decode())

    assert content == content2

    assert content2 is None


# ---------------------------------------------------------------------------
